#!/usr/bin/env python2
from scipy import *
from load import getEnergies
from numpy import max, min
from scipy import save, load

def getPaDpa(intergenic, nBases=4):
    #get the (di)nucleotide prob from the intergenic data
    pa = histogram(intergenic, bins=range(nBases+1))[0]
    pa = pa/float(sum(pa))
    dnuc = intergenic[:-1]*nBases + intergenic[1:]
    dpa = histogram(dnuc, bins=arange(nBases**2+1))[0].astype(float)
    dpac = dpa/repeat(sum(dpa.reshape(nBases,nBases),1),nBases) #transition prob
    return pa,dpac

def getSeqProb(seq, pa=None, dpac=None):
    #get probability of each sequence
    if pa == None:
        weights = None
    elif dpac == None:
        weights = product(pa[seq],1)
    else:
        #prob of a sequence "abcde" is p_a * p_ab * p_bc * p_cd ...
        #where p_a is single nuc prob, p_ab is transition prob a->b
        nBases = len(pa)
        weights = pa[seq[:,0]]*product(dpac[seq[:,:-1]*nBases + seq[:,1:]],1)

    return weights

def getAllSeq(nBases, seqLen):
    #generate list of all possible sequences for em
    seq = empty((nBases**seqLen, seqLen), dtype=int)
    for i in range(seqLen):
        seq[:,i] = tile(repeat(arange(nBases),nBases**(seqLen-i-1)),nBases**i)
    return seq

#get the energy & prob of all possible sequences generated
#for a particular em. 
def getEs(em, pa=None, dpac=None):
    seqLen, nBases = em.shape
    seq = getAllSeq(nBases, seqLen)
    weights = getSeqProb(seq, pa, dpac)    
    return getEnergies(seq,em), weights

def getEBins(em, N):
    mine = sum(min(em,1))  #store the old mine to unshift at the end
    maxe = sum(max(em,1))
    x = mine + arange(N)*(maxe-mine)/(N-1)
    dx = x[1]-x[0]
    xbins = r_[x[0]-dx/2, x + dx/2] #hist bin boundaries
    return x, dx, xbins

def getEDistExact(em,N): #XXX TODO: allow weights
    seqLen, nBases = em.shape

    x, dx, xbins = getEBins(em,N)

    splitInd = 7 #histogram chunk size
    parts = split(em, range(splitInd,em.shape[0],splitInd)) #split into chunks
    pes = [getEs(p)[0] for p in parts] #get en for each chunk
    print parts
    print [len(p) for p in pes]

    def genEs(pes):
        if pes == []:
            yield 0
        elif len(pes) == 1: #this is unnecessary, but optimization
            for x in pes[0]:
                yield x
        else:
            for x in pes[0]:
                for y in genEs(pes[1:]):
                    yield x + y

    h = zeros(len(xbins)-1,dtype=int)
    for e in genEs(pes[1:]):
        h += histogram(pes[0] + e, bins=xbins)[0]
    return x, h

def getEDist(em, bins=1024, pa=None, dpac=None, binwidth=None):
    #pa are the nucleotide probabilities, and dpac are the conditional
    #dinucleotide probabilities (the transition probabilities) in a 1d array,
    #indexed as [i*nBases + j] for transition from i->j

    #shift em so min e is 0. This is convenient for convolutions, so lower
    #end of all histograms is at E = 0 (thus no change of coords needed)
    mine = sum(min(em,1))  #store the old mine to unshift at the end
    em = em - min(em,1)[:,newaxis]
    maxe = sum(max(em,1))
    
    if binwidth == None:
        x = arange(bins)*maxe/(bins-1)
    else:
        x = arange(0, maxe, binwidth)
    dx = x[1]-x[0]
    xbins = r_[x[0]-dx/2, x + dx/2] #hist bin boundaries
    
    #split up the em into chunks of manageable size (ie, we can fit an array
    #of all possible sequences for a chunk in memory, in getEs) and get the
    #dist for each chunk. Then convolve the dists from each chunk together 
    #to get total dist.
    splitInd = 4 #histogram chunk size
    parts = split(em, range(splitInd,em.shape[0],splitInd)) #split into chunks
    ews = [getEs(p, pa, dpac) for p in parts] #get en & prob for each chunk
    part_ffts = [fft(histogram(e, weights=w, bins=xbins)[0]) for e,w in ews]
    return x + mine, ifft(reduce(multiply, part_ffts)).real

#because the chunk + convolve method has error, it is good to get the dist
#at higher resolution (smaller bin size) than needed and then resample
#into a lower reolution final dist, with smaller error per bin
def resampleDist((x,h),n=16): #x are bin centers, h are bin values
    if len(h)%n != 0: #pad w zeros
        nz = n-len(h)%n
        h = r_[h, zeros(nz)]
        x = r_[x, arange(1,nz+1)*(x[1]-x[0]) + x[-1]]

    return mean(x.reshape(len(x)/n,n),1), sum(h.reshape(len(h)/n,n),1)

def main():
    from scipy.stats.distributions import norm
    import getChi
    import pylab

    em = getChi.renormalize(norm.rvs(0,1,size=(10,4)))
    es = getEs(em)

    x,h = getEDist(em, N=1024*16)
    dx = x[1]-x[0]
    xbins = r_[x[0]-dx/2, x + dx/2]
    rh = histogram(es, bins=xbins)[0]

    h = resample((x,h))[1]
    x,rh = resample((x,rh))

    pylab.plot(x,h)
    pylab.plot(x,rh)
    pylab.show()

def test():
    import getChi, pylab
    from scipy.distributions import norm

    def nrmlz(x,f):
        return f/(sum(f)*(x[1]-x[0]))

    em = getChi.renormalize(norm.rvs(0,1,size=(7,4)))

    xh,h = getEDistExact(em, 100)
    xf,f = getEDist(em, N=100)

    pylab.plot(xh, nrmlz(xh,h))
    pylab.plot(xf, nrmlz(xf,f))
    pylab.show()

def main():
    import getChi
    import pylab, sys
    from load import loadSites

    #get the (di)nucleotide prob from the intergenic data
    nBases = 4
    intergenic = loadSites(sys.argv[2], 'ACGT')[0].flatten()
    pa,dpac = getPaDpa(intergenic)

    em = loadtxt(sys.argv[1])
    x,h = resampleDist(getEDist(em, 256*16, pa, dpac),16)
    pylab.plot(x, log10(h))
    pylab.show()

if __name__ == '__main__':
    main()
