#!/usr/bin/python2
import hashlib, os, tempfile, shutil, re
import time, threading, signal
from subprocess import Popen, PIPE
from math import ceil
#see also http://tex.stackexchange.com/questions/44486/pixel-perfect-vertical-alignment-of-image-rendered-tex-snippets
#also http://tex.stackexchange.com/questions/15604/how-to-speed-up-pdflatex-for-a-very-large-document-on-macos-x/15606#15606
#http://latexrender-ng.sourceforge.net/

staticPreambleFile = os.path.abspath('./staticpreamble')

staticpreamble = r'''
\documentclass[12pt]{article}
\usepackage[active]{preview}
\pagestyle{empty}
\usepackage[utf8]{inputenc}
\usepackage{amssymb,amsmath,amsfonts}
'''

dynamiccontent = '%&' + staticPreambleFile + r'''
\begin{document}
\newcommand{\formula}{CONTENT}
\setbox0\hbox{\formula}

\begin{preview}
\formula
\end{preview}

\newwrite\outputstream
\immediate\openout\outputstream=dimensions.dat
\immediate\write\outputstream{box_height=\the\ht0}
\immediate\write\outputstream{box_depth=\the\dp0}
\immediate\write\outputstream{box_width=\the\wd0}
\immediate\closeout\outputstream

\end{document}
'''


#optimization: precompile preamble
if not os.path.exists(staticPreambleFile + '.fmt'):
    with open(staticPreambleFile + '.tex', 'wt') as f:
        f.write(staticpreamble)
    os.system('latex -ini "&latex {} \dump"'.format(staticPreambleFile+'.tex'))
    os.remove(staticPreambleFile + '.tex')
    os.remove(staticPreambleFile + '.log')

def run_timeout(command, targetdir, timeout):
    """
    Run a sub-program in subprocess.  kill it if the specified timeout has
    passed. returns a tuple of success, stdout, stderr
    """
    kill_check = threading.Event()
    def _kill_proc_after_timeout(pid):
        os.kill(pid, signal.SIGTERM)
        kill_check.set() # tell the main routine that we had to kill
        # use SIGKILL if hard to kill...
        return
    p = Popen(command, bufsize=1, cwd=targetdir,
              stdin=PIPE, stdout=PIPE, stderr=PIPE)
    pid = p.pid
    watchdog = threading.Timer(timeout, _kill_proc_after_timeout, args=(pid, ))
    watchdog.start()
    (stdout, stderr) = p.communicate()
    watchdog.cancel() # if it's still waiting to run
    killed = kill_check.isSet()
    kill_check.clear()
    return (killed, p.returncode, stdout, stderr)

def latexhash(tex):
    return hashlib.sha1(tex.encode('utf-8', 'ignore')).hexdigest()
    
def svgfilename(texhash):
    return "latex_{}.svg".format(texhash)

def makeLatex(inputtex):        
    tex = (dynamiccontent.replace('CONTENT', inputtex)).encode('utf-8')
    texhash = latexhash(inputtex)
    fn = "latex_" + texhash

    tmpdir = tempfile.mkdtemp()
    try:
        with open(os.path.join(tmpdir, fn+'.tex'), 'wt') as f:
            f.write(tex)

        os.symlink(os.path.join(staticPreambleFile + '.fmt'), 
                   '{}/{}.fmt'.format(tmpdir, os.path.basename(staticPreambleFile)))
        
        kill,stat,out,err = run_timeout(['latex', "--interaction=nonstopmode", 
                 "-halt-on-error", "--no-shell-escape", "-parse-first-line", 
                 "{}.tex".format(fn)], tmpdir, 10)
        if kill or stat != 0:
            raise Exception('latex failed:' + out)

        getdim = lambda (n,v): (n, float(v))
        dimpat = re.compile('^(\S+)=(-?[0-9\.]+)pt')
        with open(os.path.join(tmpdir, 'dimensions.dat'), 'rt') as f:
            boxdims = dict(getdim(dimpat.match(l).groups()) for l in f)
        bdepth, bwidth, bheight = [float(boxdims[v])
                        for v in ['box_depth', 'box_width', 'box_height']]

        kill,stat,out,err = run_timeout(['dvisvgm', '--no-fonts', '--exact',
                          "{}.dvi".format(fn)], tmpdir, 10)
        if kill or stat != 0:
            raise Exception('dvisvg failed')

        with open("{}/{}.svg".format(tmpdir, fn), 'rt') as f:
            svgdat = f.read()
        svgcontent = svgdat[re.search(r'<svg [^>]*>', svgdat).end():]

        #if there are multiple svg doc fragments in a single html page
        #their ids will interfere. Therefore, each doc fragment must
        #have it own 'local' ids, which I do by prepending the tex hash
        #to all ids and refs
        uniqid = lambda m: r"id='{}'".format(texhash+m.groups()[0])
        svgcontent = re.sub(r"id='([^']+)'", uniqid, svgcontent)
        uniqref = lambda m: r"'#{}'".format(texhash+m.groups()[0])
        svgcontent = re.sub(r"'#([^']+)'", uniqref, svgcontent)
        #alternately, see if one can combine defs from all svgs in one place

        #make viewport = bbox, padded on top & bot by 1pt
        viewbox = (0, -bheight-1, bwidth, bdepth+bheight+2)

        #get svg w/h and convert from tex pt to html pt
        height = (bheight+bdepth+2)*72/72.27
        width = bwidth*72/72.27
        depth = (bdepth+1)*72/72.27

        #scale to em units (so svg will scale with font size
        scale = 1.0/8.0 #no reason why 8, it just looks good
        height, width, depth = scale*height, scale*width, scale*depth

        #XXX should I check that viewbox is larger than the viewbox generated by
        #dvisvgm with --exact, and pad if not?

        svg = '''
<svg height='{height}em' width='{width}em' 
     viewBox='{viewbox}' 
     version='1.1' style='vertical-align: -{depth}em;' 
     xmlns='http://www.w3.org/2000/svg' 
     xmlns:xlink='http://www.w3.org/1999/xlink'>{svg}'''.format(svg=svgcontent,
                                    height=height, width=width, depth=depth,
                                    viewbox=" ".join(str(x) for x in viewbox))

    finally:
        shutil.rmtree(tmpdir)

    return svg

if __name__ == '__main__':
    print makeLatex(r'''$E = m c^2$''', './')



#used to debug baseline position:

            #svgcontent = svgcontent[:-12] + '<rect x="0" y="0" width="1" height="1" fill="red"/><rect x="0" y="1" width="1" height="1" fill="blue"/></g></svg>'
    #return '<span style="display: inline-block; width: 5px; height: 5px; background-color: green; vertical-align: baseline;"></span>' + svg


#<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
    #    svg = '''<svg height='{height}em' width='{width}em' viewBox='{viewbox}' 
     #version='1.1' style='vertical-align: -{depth}em;' 
     #xmlns='http://www.w3.org/2000/svg' preserveAspectRatio="none"
